
//
//
// Project       : Vorpal-Mail
//
// Source        : array.cc
//
// Version       : 1.0
//
// Creator       : Blane Bramble                    Date          : 27th October 2000
// Amended       : Blane Bramble                    Date          : 23rd January 2001
//
// License       : Copyright Virtua-Web 2000-2001, licensed under GNU General Public License (GPL) v2
//
// Description   :
//
// Class to support extendible arrays (initially for strings). This is needed
// by the configuration file/virtual domain support for lists of users etc.
//
//
// Changelog     :
//
// 7 Nov 2000
// Added touch routines to allow a search routine to exclude array items from further
// searches if they have already been matched (to prevent circular references in
// users and lists etc).
//
// 23rd Jan 2001
// Altered to use new STRING class instead of char []

#include <stdio.h>
#include <stdlib.h>

#include "array.h"

STRING_ARRAY::STRING_ARRAY(void)
{
array = 0;
size = 0;
}

STRING_ARRAY::~STRING_ARRAY()
{
int n;
delete[] array;
}

int STRING_ARRAY::add(char *s)
{
STRING *old;
int n;
old = array;
array = new STRING[size+1];
for(n=0;n<size;n++)
	array[n] = old[n];
delete[] old;
array[size].set(s);
return size++;
}

int STRING_ARRAY::find(char *s)
{
int n;
for(n=0;n<size;n++)
	if (array[n].compare(s) == 0)
		return n;

return -1;
}

char *STRING_ARRAY::get(int n)
{
if (size == 0 || array == NULL || n >= size)
	return NULL;
return array[n].get();
}

void STRING_ARRAY::remove(int n)
{
STRING *newarray;

if (size > 0)
	{
	int i;

	if (size == 1)
		{
		delete[] array;
		array = 0;
		size = 0;
		}
	else
		{
		newarray = new STRING[size-1];

		for(i=0;i<n;i++)
			newarray[i] = array[i];

		for(i=n+1;i<size;i++)
			newarray[i-1] = array[i];

		delete[] array;
		array = newarray;
		size--;
		}
	}
}

void STRING_ARRAY::replace(int n, char *s)
{
if (n < size)
	array[n].set(s);
}

