
//
//
// Project       : Vorpal-Mail
//
// Source        : queue.cc
//
// Version       : 1.0
//
// Creator       : Blane Bramble                    Date          : 22nd February 2001
// Amended       : Blane Bramble                    Date          : 27th March 2001
//
// License       : Copyright Virtua-Web 2000-2001, licensed under GNU General Public License (GPL) v2
//
// Description   :
//
// Definitions for Queue class to manage a message queue and handle priorities
// etc.
//
// Changelog     :
//

#include <stdio.h>
#include <stdlib.h>

#include <sys/stat.h>
#include <unistd.h>

#include <limits.h>
#include <string.h>

#include "queue.h"

QUEUE::QUEUE(void)
{
id.set("");
from.set("");
result.set("");
qtime = 0; ltime = 0;
size = 0;
}

QUEUE::QUEUE(char *file)
{

this->load(file);
}

QUEUE::~QUEUE()
{
}

long QUEUE::priority(void) const
{
time_t now;
long age = 0,nsize,bits=1;
now = time(NULL);
if (qtime != 0)
	age = (now - qtime)/60;
nsize = size/1024;		// size in K
while(nsize > 0)
	{
	nsize >>= 1;
	bits++;
	}
return (age / bits);
}

void QUEUE::load(char *file)
{
char path[PATH_MAX], fname[PATH_MAX], line[256];
struct stat finfo;
FILE *fp;

id.set(file);

getcwd(path, PATH_MAX);
snprintf(fname, PATH_MAX, "%s/R%s", path, file);
stat(fname, &finfo);

size = finfo.st_size;

snprintf(fname, PATH_MAX, "%s/H%s", path, file);
if ((fp = fopen(fname, "r")) != NULL)
	{
	while(fgets(line, 255, fp) != NULL)
		{
		char *eol;
		while((eol = strchr(line, '\n')) != NULL)
			*eol = 0;
		while((eol = strchr(line, '\r')) != NULL)
			*eol = 0;
		switch(*line)
			{
			case 'B':
				bouncecount = strtol(line+1, NULL, 10);
				break;
			case 'F':
				from.set(line+1);
				break;
			case 'Q':
				qtime = strtol(line+1, NULL, 10);
				break;
			case 'L':
				ltime = strtol(line+1, NULL, 10);
				break;
			case 'R':
				result.set(line+1);
				break;
			case 'T':
				to.add(NULL, new STRING(line+1));
				break;
			case 'D':
				delivered.add(NULL, new STRING(line+1));
				break;
			}
		}
	fclose(fp);
	}
}

int QUEUE::operator<(const QUEUE &right) const
{
return this->priority() < right.priority();
}

int QUEUE::operator>(const QUEUE &right) const
{
return this->priority() > right.priority();
}

int QUEUE::operator==(const QUEUE &right) const
{
return priority() == right.priority();
}

