
//
//
// Project       : Vorpal-Mail
//
// Source        : smtp-client.cc
//
// Version       : 1.0
//
// Creator       : Blane Bramble                    Date          : 16th March 2001
// Amended       : Blane Bramble                    Date          : 27th March 2001
//
// License       : Copyright Virtua-Web 2000-2001, licensed under GNU General Public License (GPL) v2
//
// Description   :
//
// Derived SMTPCLIENT class (from SMTP main class). Provides the necessary code
// to create an SMTP client (sending) session.
//
// Changelog     :
//

#include <iostream.h>
#include <stdio.h>
#include <stdlib.h>
#include <pwd.h>
#include <stdarg.h>
#include <unistd.h>
#include <string.h>
#include <time.h>
#include <ctype.h>
#include <syslog.h>
#include <errno.h>
#include <fcntl.h>

#include <netdb.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <sys/wait.h>


#include "mail.h"
#include "array.h"
#include "smtp-client.h"
#include "smtp-server.h"
#include "virtual.h"

SMTPCLIENT::SMTPCLIENT(void)
{
initialise();
}

SMTPCLIENT::SMTPCLIENT(int s, struct sockaddr_in *sa)
{
initialise();
memcpy(&this->saddr, sa, sizeof(struct sockaddr_in));
socket = s;
}

int SMTPCLIENT::openfiles(char *fn)
{
filename = fn;
	
if (bodyfile == NULL)
	bodyfile = fopen(makefilename(domain->getqout(), 'R', fn), "r+");	

if (envelopefile == NULL)
	envelopefile = fopen(makefilename(domain->getqout(), 'F', fn), "r+");

if (bouncefile == NULL)
	bouncefile = fopen(makefilename(domain->getqout(), 'B', fn), "w");

if (bodyfile == NULL || envelopefile == NULL || bouncefile == NULL)
	return 0;
		
lock(bodyfile);
lock(envelopefile);
return 1;
}

void SMTPCLIENT::initialise(void)
{
SMTP_SESSION::initialise();
}

void SMTPCLIENT::process_client(char *data)
{
int code;

code = *data;
if (SMTP_SESSION::debug > 2 || (SMTP_SESSION::debug > 1 && this->status != SMTP_STATUS_DATASEND))
	syslog(LOG_MAIL + LOG_DEBUG, "<< %s\n", data);

last_response = data;

switch(this->status)
	{
	case SMTP_STATUS_OPEN:
		switch(code)
			{
			case '2':
				clientstatus(SMTP_STATUS_HELO);
				break;
			default:
				this->status = SMTP_STATUS_ABORT;
				break;
			}
		break;
	case SMTP_STATUS_RSET:
		switch(code)
			{
			case '2':
				clientstatus(SMTP_STATUS_MAIL, from.get());
				break;
			default:
				this->status = SMTP_STATUS_ABORT;
				break;
			}
		break;
	case SMTP_STATUS_HELO:
		switch(code)
			{
			case '2':
				clientstatus(SMTP_STATUS_MAIL, from.get());
				break;
			default:
				this->status = SMTP_STATUS_ABORT;
				break;
			}
		break;
	case SMTP_STATUS_MAIL:
		switch(code)
			{
			case '2':
				if (to.get(0) == 0)
					this->status = SMTP_STATUS_ABORT;
				else
					{
					this->loop = 1;
					clientstatus(SMTP_STATUS_RCPT, to.get(0));
					}
				break;
			default:
				this->status = SMTP_STATUS_ABORT;
			}
		break;
	case SMTP_STATUS_RCPT:
		switch(code)
			{
			case '2':		// add previous to pending
				pending.add(to.get(this->loop-1));
				to.remove(this->loop-1);
				if (to.get(this->loop) != NULL)
					clientstatus(SMTP_STATUS_RCPT, to.get(this->loop));
				else
					clientstatus(SMTP_STATUS_DATA);
				break;
			default:
				this->loop++;
				if (to.get(this->loop) != NULL)
					clientstatus(SMTP_STATUS_RCPT, to.get(this->loop));
				else
					clientstatus(SMTP_STATUS_DATA);
				break;
					
			}
		break;
	case SMTP_STATUS_DATA:
		switch(code)
			{
			case '3':
				if (this->envelopefile == NULL || this->bodyfile == NULL)
					this->status = SMTP_STATUS_ABORT;
				else
					this->status = SMTP_STATUS_DATASEND;
				break;
			default:
				this->status = SMTP_STATUS_ABORT;
				break;
			}
		break;
	case SMTP_STATUS_DATASEND:		// need to move logic to send a line at a time and return to here
	case SMTP_STATUS_SENDRETRY:
		if (this->envelopefile != NULL)
			{
			if (status == SMTP_STATUS_SENDRETRY || fgets(spill, SMTP_MAX_LINE, this->envelopefile) != NULL)
				{
				char *ptr;
				if ((ptr = strchr(spill, '\n')) != NULL)
					*ptr = 0;
				if ((ptr = strchr(spill, '\r')) != NULL)
					*ptr = 0;
				clientstatus(SMTP_STATUS_DATASEND, spill);
				}
			else
				{
				unlock(envelopefile);
				fclose(this->envelopefile);
				this->envelopefile = NULL;
				}
			}
		else if (this->bodyfile != NULL)
			{
			if (status == SMTP_STATUS_SENDRETRY || fgets(spill, SMTP_MAX_LINE, this->bodyfile) != NULL)
				{
				char *ptr;
				if ((ptr = strchr(spill, '\n')) != NULL)
					*ptr = 0;
				if ((ptr = strchr(spill, '\r')) != NULL)
					*ptr = 0;
				clientstatus(SMTP_STATUS_DATASEND, spill);
				}
			else
				{
				clientstatus(SMTP_STATUS_DATASEND, NULL);
				unlock(bodyfile);
				fclose(this->bodyfile);
				this->bodyfile = NULL;
				this->status = SMTP_STATUS_SENT;
				}
			}
		break;
	case SMTP_STATUS_DATATEXT:
		clientstatus(SMTP_STATUS_DATASEND, NULL);
		break;
	case SMTP_STATUS_SENT:
		switch(code)
			{
			case '2':
				this->status = SMTP_STATUS_COMPLETE;
				cleanup(NULL);			// remove files
				break;
			default:
				this->status = SMTP_STATUS_ABORT;
				break;
			}
		break;
	}

if (status == SMTP_STATUS_ABORT || status == SMTP_STATUS_COMPLETE)
	{
	ltime = time(NULL);

	while(pending.getsize() > 0)		// move pending to delivered or back to to
		{
		if (status == SMTP_STATUS_COMPLETE)
			delivered.add(pending.get(0));
		else
			to.add(pending.get(0));
		pending.remove(0);
		}

	if (status == SMTP_STATUS_ABORT || to.getsize() > 0)
		{
		if ((ltime - qtime) < domain->getbounceafter())
			{
			int n = bouncecount;
			writeheader();
			bouncecount = domain->checkbouncecount(ltime-qtime, bouncecount);

			if (n != bouncecount)
				{
				SMTPSERVER bouncer;
				int n;
				bouncer.setfrom(from.get());
				for(n=0;n<to.getsize();n++)
					bouncer.addto(to.get(n));
				addbounce("Your message has not been delivered to all recipients after %ld days\n", bouncecount);
				addbounce("You do not need to do anything. I will continue to try and send this message\n");
				addbounce("until it's expiry time is exceeded. I will inform you if this occurs.\n");
				addbounce("Last error reported: %s\n", last_response.get());
				bouncer.sendbounce(filename.get());
				while(to.getsize() > 0)
					to.remove(0);
				}
			}
		else
			{
			SMTPSERVER bouncer;
			int n;

			bouncer.setfrom(from.get());
			for(n=0;n<to.getsize();n++)
				bouncer.addto(to.get(n));

			addbounce("Could not deliver message to all recipients after %ld days\n", (long)(domain->getbounceafter() / (60 * 60 * 24)));
			addbounce("** This is a permanent error. Your message has not been delivered **\n");
			addbounce("Last error reported: %s\n", last_response.get());
			bouncer.sendbounce(filename.get());

			while(to.getsize() > 0)
				to.remove(0);
			}
		}
	}
}

void SMTPCLIENT::clientstatus(SMTP_STATUS stat, char *str)
{
switch(stat)
	{
	case SMTP_STATUS_TIMEOUT:
		this->status = SMTP_STATUS_ABORT;
		ltime = time(NULL);
		writeheader();
		break;
	case SMTP_STATUS_HELO:
		if (sendf("HELO %s\r\n", hostname))
			this->status = SMTP_STATUS_HELO;
		else
			this->status = SMTP_STATUS_ABORT;
		break;
	case SMTP_STATUS_MAIL:
		if (sendf("MAIL FROM:<%s>\r\n", str))
			this->status = SMTP_STATUS_MAIL;
		else
			this->status = SMTP_STATUS_ABORT;
		break;
	case SMTP_STATUS_RCPT:
		if (sendf("RCPT TO:<%s>\r\n", str))
			this->status = SMTP_STATUS_RCPT;
		else
			this->status = SMTP_STATUS_ABORT;
		break;
	case SMTP_STATUS_DATA:
		if (sendf("DATA\r\n"))
			this->status = SMTP_STATUS_DATA;
		else
			this->status = SMTP_STATUS_ABORT;
		break;
	case SMTP_STATUS_DATASEND:
		this->status = SMTP_STATUS_DATASEND;
		if (str == NULL)
			{
			if (sendf(".\r\n"))
				this->status = SMTP_STATUS_SENT;
			else
				this->status = SMTP_STATUS_SENDRETRY;
			}
		else if (*str == '.')
			{
			if (sendf(".%s\r\n", str))
				this->status = SMTP_STATUS_DATASEND;
			else
				this->status = SMTP_STATUS_SENDRETRY;
			}
		else
			{
			if (sendf("%s\r\n", str))
				this->status = SMTP_STATUS_DATASEND;
			else
				this->status = SMTP_STATUS_SENDRETRY;
			}
		break;
	}
}

void SMTPCLIENT::clientstatus(SMTP_STATUS stat)
{
clientstatus(stat, "[none]");
}

void SMTPCLIENT::pulse(void)
{
if (timeout)
	{
	if (--timeout == 0)
		clientstatus(SMTP_STATUS_TIMEOUT);
	}
}

void SMTPCLIENT::process(char *data)
{
char *from,*to,cmd[SMTP_MAX_LINE];

from = data;
strcpy(cmd, spill);

to = (cmd + strlen(cmd));

while(*from != 0)
	{
	if (*from == '\r')
		{
		*to = 0;
		process_client(cmd);
		to = cmd;
		if (*(from+1) == '\n')
			from++;
		}
	else
		*to++ = *from;
	from++;
	}
*to = 0;
strcpy(spill, cmd);		// buffer partial command back
}

void SMTPCLIENT::sendquit(void)
{
if (sendf("QUIT\r\n"))
	this->status = SMTP_STATUS_COMPLETE;
else
	this->status = SMTP_STATUS_ABORT;
}

void SMTPCLIENT::sendreset(void)
{
if (sendf("RSET\r\n"))
	this->status = SMTP_STATUS_RSET;
else
	this->status = SMTP_STATUS_ABORT;
}

void SMTPCLIENT::cleanup(char *s)
{
if (s == NULL)
	s = filename.get();

if (status != SMTP_STATUS_COMPLETE)
	return;

unlink(makefilename(domain->getqout(), 'R', s));
unlink(makefilename(domain->getqout(), 'F', s));
unlink(makefilename(domain->getqout(), 'H', s));
unlink(makefilename(domain->getqout(), 'B', s));
}


