
//
//
// Project       : Vorpal-Mail
//
// Source        : smtp-server.cc
//
// Version       : 1.0
//
// Creator       : Blane Bramble                    Date          : 23rd October 2000
// Amended       : Blane Bramble                    Date          : 27th March 2001
//
// License       : Copyright Virtua-Web 2000-2001, licensed under GNU General Public License (GPL) v2
//
// Description   :
//
// Class implementation for SMTP session. Implements the server side of an RFC821 compliant
// SMTP server.
//
// Changelog     :
//
// 24th Jan 2001
// Updated relay code to use new TLIST style relaylist class.
//
// 12th March 2001
// Added basic spam trap code
//
// 19th March 2001
// Renamed to smtp-server.cc as part of split of SMTP code into two seperate derived classes

#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <unistd.h>
#include <string.h>
#include <time.h>
#include <ctype.h>
#include <syslog.h>

#include <netdb.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>

#include <pwd.h>

#include "mail.h"
#include "array.h"
#include "smtp-server.h"
#include "virtual.h"

SMTPSERVER::SMTPSERVER(void)
{
initialise();
}

SMTPSERVER::SMTPSERVER(int s, struct sockaddr_in *sa)
{
initialise();
memcpy(&this->saddr, sa, sizeof(struct sockaddr_in));
socket = s;
status = SMTP_STATUS_HELO;
}

SMTPSERVER::~SMTPSERVER()
{
}

void SMTPSERVER::pulse(void)
{
if (timeout)
	{
	if (--timeout == 0)
		serverstatus(SMTP_STATUS_TIMEOUT);
	}
}

void SMTPSERVER::process(char *data)
{
char *from,*to,cmd[SMTP_MAX_LINE];

from = data;
strcpy(cmd, spill);

to = (cmd + strlen(cmd));

while(*from != 0)
	{
	if (*from == '\r')
		{
		*to = 0;
		process_server(cmd);
		to = cmd;
		if (*(from+1) == '\n')
			from++;
		}
	else
		*to++ = *from;
	from++;
	}
*to = 0;
strcpy(spill, cmd);		// buffer partial command back
}

int SMTPSERVER::openfiles(char *fn)
{
filename = fn;
	
if (bodyfile == NULL)
	bodyfile = fopen(makefilename(domain->getqin(), 'M', fn), "w");	

if (envelopefile == NULL)
	envelopefile = fopen(makefilename(domain->getqin(), 'E', fn), "w");

if (bouncefile == NULL)
	bouncefile = fopen(makefilename(domain->getqin(), 'B', fn), "w");

if (bodyfile == NULL || bouncefile == NULL || envelopefile == NULL)
	return 0;
		
lock(bodyfile);
lock(bouncefile);
lock(envelopefile);

return 1;
}

void SMTPSERVER::initialise(void)
{
SMTP_SESSION::initialise();

relayok = 0;
}

void SMTPSERVER::reset(void)
{
SMTP_SESSION::reset();

cleanup(NULL);

bounce = 0;
}

void SMTPSERVER::cleanup(char *s)
{
if (s == NULL)
	s = filename.get();

unlink(makefilename(domain->getqin(), 'M', s));
unlink(makefilename(domain->getqin(), 'E', s));
unlink(makefilename(domain->getqin(), 'B', s));
unlink(makefilename(domain->getqin(), 'H', s));
}

void SMTPSERVER::virusmsg(char *log)
{
SMTPSERVER virus;
char tmp[SMTP_MAX_NAME];
TLIST<STRING> *local;

//
// Add the expanded FROM list to our virus msg recipients

virus.addto(from.get());

//if (domain != NULL && (local = domain->expand(from.get())) != NULL)
//	{
//	TWALK<STRING> walk;
//	walk.attach(local);
//	if (walk.start())
//	do	{
//		virus.addto(walk.getkey());
//		} while(walk.next());
//	delete local;
//	}
//
// Set the from the postmaster@ourdomain
//
snprintf(tmp, SMTP_MAX_NAME-1, "postmaster@%s", domain->getname());
virus.from = tmp;

virus.filename = virus.sessionname();
virus.virushdr(filename.get(), log);

virus.localdeliver();
virus.writeheader();
virus.remotedeliver();

virus.cleanup(NULL);

closefiles();
status = SMTP_STATUS_COMPLETE;
cleanup(NULL);

#ifdef OLDVIRUSMETHOD
// Old code

msgto = from;
session = filename;

closefiles();						// close old file
status = SMTP_STATUS_COMPLETE;

filename = sessionname();			// grab a new session ID
virushdr(session.get(), log);

while(to.getsize() > 0)
	to.remove(0);

if (domain != NULL && (local = domain->expand(msgto.get())) != NULL)		// expand and convert to to user
	{
	TWALK<STRING> walk;
	walk.attach(local);
	if (walk.start())
	do	{
		addto(walk.getkey());
		} while(walk.next());
	delete local;
	}
else
	addto(msgto.get());
	
snprintf(tmp, SMTP_MAX_NAME-1, "postmaster@%s", domain->getname());
from = tmp;

localdeliver();
writeheader();
remotedeliver();

cleanup(filename.get());

// TODO: lots!

filename = session;
#endif
}

int SMTPSERVER::scanvirus(char *log)
{
char line[MAX_CMD];
char tmpname[] = "/tmp/vmailtmpXXXXXX";
int fd, pid;
if ((fd = mkstemp(tmpname)) == -1)
	syslog(LOG_MAIL + LOG_DEBUG, "** Could not make temporary file, virus scan aborted");
else
	{
	FILE *out;

	snprintf(log, 30, "%s.log", tmpname);

	if ((out = fdopen(fd, "w")) == NULL)
		{
		syslog(LOG_MAIL + LOG_DEBUG, "** Could not open temporary file, virus scan aborted");
		close(fd);
		}
	else
		{
		FILE *in;
		int status;

		if ((in = fopen(makefilename(domain->getqin(),'E',filename.get()), "r+")) == NULL)
			{
			syslog(LOG_MAIL + LOG_DEBUG, "** Could not open envelope file, virus scan aborted");
			fclose(out);
			return 0;
			}

		lock(in);	
		while(fgets(line, MAX_CMD-1, in) != NULL)
			fputs(line, out);
		unlock(in);
		fclose(in);

		fputs("\n", out);		// blank line seperates

		if ((in = fopen(makefilename(domain->getqin(),'M',filename.get()), "r+")) == NULL)
			{
			syslog(LOG_MAIL + LOG_DEBUG, "** Could not open message file, virus scan aborted");
			fclose(out);
			return 0;
			}
		
		lock(in);	
		while(fgets(line, MAX_CMD-1, in) != NULL)
			fputs(line, out);
		unlock(in);
		fclose(in);

		fclose(out);

		if (debug)
			syslog(LOG_MAIL + LOG_DEBUG, "Virus scanner started (tmp file %s, session %s)", tmpname, filename.get());

		switch((pid = fork()))
			{
			case -1:
				syslog(LOG_MAIL + LOG_DEBUG, "** Could not fork, virus scan aborted");
				break;
			case 0:		// child!
				if (debug)
					syslog(LOG_MAIL + LOG_DEBUG, "Exec: %s %s", domain->getscanner(), tmpname+5);
				execl(domain->getscanner(), domain->getscanner(), tmpname + 5, log, NULL);
				syslog(LOG_MAIL + LOG_DEBUG, "Exec failed, %d (%s)", errno, strerror(errno));
				_exit(1);
				break;
			default:
				if (waitpid(pid, &status, 0) != pid)
					syslog(LOG_MAIL + LOG_DEBUG, "** Error waiting for scanner to exit, virus scan aborted");
				else
					{
					if (debug)
						syslog(LOG_MAIL + LOG_DEBUG, "Virus scanner finished");
					if ((in = fopen(tmpname, "r+")) == NULL)		// file inaccessable!
						return 1;
					else
						{
						fclose(in);
						unlink(tmpname);		// delete temporary file
						addheader("X-Vorpal", "Virus Scanner " PROGRAM "\n");
						return 0;
						}
					}
				break;
			}
		}
	}
return 0;
}

void SMTPSERVER::virushdr(char *sess, char *log)
{
FILE *msgfile, *data;
char tm[SMTP_MAX_LINE];
time_t t;
int n;

time(&t);
strftime(tm, sizeof(tm), "%a, %d %b %Y %H:%M:%S %z", gmtime(&t));

if ((msgfile = fopen(makefilename(domain->getqin(), 'E', filename.get()), "w")) == NULL)
	return;

lock(msgfile);

fprintf(msgfile, "Date: %s\n", tm);
fprintf(msgfile, "From: @%s\n", from.get());

for(n=0;n<to.getsize();n++)
	fprintf(msgfile, "%s %s\n", (n==0) ? "To:" : "", to.get(n));

fprintf(msgfile, "Subject: ** Virus Warning **\n");

unlock(msgfile);
fclose(msgfile);

if ((msgfile = fopen(makefilename(domain->getqin(), 'M', filename.get()), "w")) == NULL)
	return;

lock(msgfile);	

fprintf(msgfile, "This is an automatically generated message. Do not reply to it.\n\n");
fprintf(msgfile, "A message sent by you has been received and was found to contain a virus.\n\n");
fprintf(msgfile, "As a result this message has not been delivered to it's recipient(s).\n\n");
fprintf(msgfile, "The result of the virus scan, and your original message headers are\n");
fprintf(msgfile, "reproduced below for your information.\n\n");

fprintf(msgfile, "------ Scanner transscript ------\n\n");

if ((data = fopen(log, "r+")) == NULL)
	fprintf(msgfile, "    [ ERROR: Transcript lost ]\n");
else
	{
	
	lock(data);
	
	while(fgets(tm, SMTP_MAX_LINE-1, data)!= NULL)
		fprintf(msgfile, "    %s", tm);
		
	unlock(data);
		
	fclose(data);
	}

fprintf(msgfile, "\n------ Message Headers ------\n\n");

if ((data = fopen(makefilename(domain->getqin(), 'E', sess), "r+")) == NULL)
	fprintf(msgfile, "    [ ERROR: Message Headers lost ]\n");
else
	{
	
	lock(data);
	
	while(fgets(tm, SMTP_MAX_LINE-1, data)!= NULL)
		fprintf(msgfile, "    %s", tm);
		
	unlock(data);		
		
	fclose(data);
	}

unlock(msgfile);	
	
fclose(msgfile);
}

void SMTPSERVER::process_data(char *data)
{
if (SMTP_SESSION::debug > 2)
	syslog(LOG_MAIL + LOG_DEBUG, "<< %s\n", data);

status = SMTP_STATUS_DATA;

if (strcmp(data, ".") == 0)
	{
	int pid;
	char virlog[40];

	if (qtime == 0)
		qtime = time(NULL);

	check_header();
	unlock(envelopefile);
	fclose(envelopefile);
	envelopefile = NULL;

	unlock(bodyfile);
	fclose(bodyfile);
 	bodyfile = NULL;
       	serverstatus(250);
	switch((pid = fork()))
		{

		case -1:
			syslog(LOG_MAIL + LOG_DEBUG, "** Could not fork. Delivering %s in same process", filename.get());
		case 0:
			if (scanvirus(virlog))
				{
				syslog(LOG_MAIL + LOG_DEBUG, "** Virus found in message %s", filename.get());
				closefiles();
				virusmsg(virlog);
				unlink(virlog);
				cleanup(NULL);
				}
			else
				{
				unlink(virlog);
				deliver();
				}
			if (pid == 0)
				_exit(0);
			break;
		default:
			if (debug)
				syslog(LOG_MAIL + LOG_DEBUG, "Child %d is delivering mail", pid);
			break;
		}
	status = SMTP_STATUS_MAIL;
	}
else
	{
	if (dataheader)
		{
		char head[SMTP_MAX_LINE], tail[SMTP_MAX_LINE];
		if (!isspace(*data) && (!splitheader(data, head, tail) || strchr(head, ' ')!= NULL))		// not a header!
			{
			dataheader = 0;
			if (*data != 0)
				fprintf(bodyfile, "%s\n", data);
			}
		else
			{
			if (isspace(*data))
				{
				if (lastheader != NULL && lastheader->getdata() != NULL)
					lastheader->getdata()->add(data);
				}
			else
				lastheader = header.add(head, new STRING(tail));
			}
		}
	else
		{
		if (bodyfile != NULL)
			{
			if (*data == '.')
				fprintf(bodyfile, "%s\n", data+1);
			else
				fprintf(bodyfile, "%s\n", data);
			}
		else
			{
			addbounce("Internal error: could not spool message.\n");
			serverstatus(451);
			sendbounce(filename.get());
			status = SMTP_STATUS_MAIL;
			return;
			}
		}
	
	}
}

void SMTPSERVER::check_header(void)
{
STRING *thishead;
TWALK<STRING> headwalk;

if (header.gethead() == 0)
	return;

if (envelopefile == NULL)
	{
	addbounce("Internal error: could not spool envelope.\n");
	serverstatus(451);
	sendbounce(filename.get());
	status = SMTP_STATUS_MAIL;
	return;
	}


if ((thishead = header.find("MIME-Version")) != NULL)
	{
	if (strcasecmp(thishead->get(), "1.0") == 0)
		datamime = 1;
	}

if ((thishead = header.find("From")) == NULL)
	{
	snprintf(scratch, SMTP_MAX_LINE, "\"%s\" <%s>", from.get(), from.get());
	header.add("From", new STRING(scratch));
	}

headwalk.attach(&header);
if (headwalk.start())
	{
	do	{
		char *from, *to;
		int first = 1;


		from = headwalk.get()->get();

//		from = (char *)thishead->getdata();
// need to split into chunks if longer than an SMTP line

		while(strlen(from) > SMTP_BREAK_LINE)
			{
			int n = SMTP_BREAK_LINE;
			while(n > 0 && !isspace(*(from + n)))
				n--;
			if (n == 0)
				n = SMTP_BREAK_LINE;
			strncpy(scratch, from, n);
			*(scratch + n) = 0;
			while(isspace(*(from + n)))
				n++;
			if (first)
				fprintf(envelopefile, "%s: %s\n", headwalk.getkey(), scratch);
			else
				fprintf(envelopefile, "    %s\n", scratch);
	
			from += n;
			first = 0;
			}
		if (*from != 0)
			{
			if (first)
				fprintf(envelopefile, "%s: %s\n", headwalk.getkey(), from);
			else
				fprintf(envelopefile, "    %s\n", from);
			}
		} while(headwalk.next());
	}
}

void SMTPSERVER::sendbounce(char *sess)
{
SMTPSERVER bouncer;
char tmp[SMTP_MAX_NAME];
TLIST<STRING> *local;

snprintf(tmp, SMTP_MAX_NAME-1, "postmaster@%s", domain->getname());
bouncer.from = tmp;
bouncer.addto(from.get());

bouncer.bouncing = 1; bouncer.bounce = 0;

bouncer.filename = bouncer.sessionname();

bouncer.errormail(sess);

bouncer.localdeliver();
bouncer.writeheader();
bouncer.remotedeliver();

bouncer.cleanup(NULL);

  
#ifdef OLDBOUNCEMETHOD

bouncing = 1; bounce = 0;				// stop bounce msg from being bounced internally

msgto = from;
session = filename;

closefiles();						// close old file
status = SMTP_STATUS_COMPLETE;

filename = sessionname();			// grab a new session ID
errormail(session.get());					// create a new email

while(to.getsize() > 0)
	to.remove(0);

if (domain != NULL && (local = domain->expand(msgto.get())) != NULL)		// expand and convert to to user
	{
	TWALK<STRING> walk;
	walk.attach(local);
	if (walk.start())
		{
		do	{
			addto(walk.getkey());
			} while(walk.next());
		}
	delete local;
	}
else
	addto(msgto.get());
	
snprintf(tmp, SMTP_MAX_NAME-1, "postmaster@%s", domain->getname());
from = tmp;

localdeliver();
writeheader();
remotedeliver();

cleanup(filename.get());

// TODO: lots!

filename = session;
#endif
}

void SMTPSERVER::errormail(char *sess)
{
FILE *msgfile, *transcript, *header;
char *ptr, tm[SMTP_MAX_LINE];
time_t t;
int n;
VIRTUAL_DOMAIN *dom;

time(&t);
strftime(tm, sizeof(tm), "%a, %d %b %Y %H:%M:%S %z", gmtime(&t));

if ((msgfile = fopen(makefilename(domain->getqin(), 'E', filename.get()), "w")) == NULL)
	return;

lock(msgfile);	

//
// Firstly create a simple envelope file
//

fprintf(msgfile, "Date: %s\n", tm);
fprintf(msgfile, "From: %s\n", from.get());
for(n=0;n<to.getsize();n++)
	fprintf(msgfile, "%s %s\n", (n==0) ? "To:" : "", to.get(n));
fprintf(msgfile, "Subject: Error unable to deliver message\n");

unlock(msgfile);

fclose(msgfile);

//
// Now create the message body
//

if ((msgfile = fopen(makefilename(domain->getqin(), 'M', filename.get()), "w")) == NULL)
	return;

lock(msgfile);	
	
fprintf(msgfile, "An error occurred whilst trying to deliver a message from you.\n\n");
fprintf(msgfile, "As a result this message has not been delivered to all the recipients.\n\n");


transcript = fopen(makefilename(domain->getqin(), 'B', sess), "r+");
header = fopen(makefilename(domain->getqin(), 'E', sess), "r+");


//	transcript = fopen(makefilename(domain->getqout(), 'B', sess), "r+");
//	header = fopen(makefilename(domain->getqout(), 'F', sess), "r+");

fprintf(msgfile, "------ Error transscript ------\n\n");

if (transcript == NULL)
	fprintf(msgfile, "    [ ERROR: Transcript lost ]\n");
else
	{
	lock(transcript);
	
	while(fgets(tm, SMTP_MAX_LINE-1, transcript)!= NULL)
		fprintf(msgfile, "    %s", tm);
		
	unlock(transcript);
		
	fclose(transcript);
	}

fprintf(msgfile, "\n------ Message Headers ------\n\n");

if (header == NULL)
	fprintf(msgfile, "    [ ERROR: Message Headers lost ]\n");
else
	{
	
	lock(header);
	
	while(fgets(tm, SMTP_MAX_LINE-1, header)!= NULL)
		fprintf(msgfile, "    %s", tm);
		
	unlock(header);		
		
	fclose(header);
	}

unlock(msgfile);	
	
fclose(msgfile);
}

int SMTPSERVER::dolocaldeliver(char *to, char *sess, VIRTUAL_DOMAIN *dom)
{
char *cmdline, *ptr;
char tmpcmd[MAX_CMD];
FILE *data, *proc;
int pstat;

ptr = tmpcmd;
if ((cmdline = dom->getlocal()) == NULL)
	return 1;

while(*cmdline != 0)
	{
	if (*cmdline == '%')
		{
		switch(*(++cmdline))
			{
			default:
				break;
			case 'h':
				strcpy(ptr, domain->getname());
				ptr += strlen(ptr);
				cmdline++;
				break;
			case 'u':
				strcpy(ptr, to);
				ptr += strlen(ptr);
				cmdline++;
				break;
			case 'f':
				from = ptr;
				ptr += strlen(ptr);
				cmdline++;
				break;
			}
		}
	else
		*ptr++ = *cmdline++;
	}
*ptr = 0;

if (debug)
	syslog(LOG_MAIL + LOG_DEBUG, "Local delivery with %s", tmpcmd);

fflush(NULL);

if ((proc = popen(tmpcmd, "w")) == NULL)			// TODO: need to add error checking
	{
	syslog(LOG_MAIL + LOG_DEBUG, "** Could not run local delivery program");
	return 1;
	}

if ((data = fopen(makefilename(domain->getqin(),'E',sess), "r+")) == NULL)
	{
	syslog(LOG_MAIL + LOG_DEBUG, "** Could not open envelope for local delivery");
	return 1;
	}
	
lock(data);	
	
while(fgets(tmpcmd, MAX_CMD-1, data) != NULL)
	fputs(tmpcmd, proc);

unlock(data);

fclose(data);

fputs("\n", proc);		// add blank line to seperate headers

if ((data = fopen(makefilename(domain->getqin(),'M',sess), "r+")) == NULL)
	{
	syslog(LOG_MAIL + LOG_DEBUG, "** Could not open message body for local delivery");
	return 1;
	}
lock(data);	
	
while(fgets(tmpcmd, MAX_CMD-1, data) != NULL)
	fputs(tmpcmd, proc);

unlock(data);

fclose(data);

if (WIFEXITED((pstat = pclose(proc))))
	return WEXITSTATUS(pstat);
return -1;
}

void SMTPSERVER::localdeliver(void)
{
static char tmp[SMTP_MAX_NAME];
int n = 0;
char *msgto, *ptr;
VIRTUAL_DOMAIN *dom = domain;

while((msgto = to.get(n)) != NULL)
	{
	strcpy(tmp, msgto);
	msgto = tmp;

	ptr = strchr(msgto, '@');
	if (ptr == NULL || (domain != NULL && (dom = domain->finddomain(ptr+1)) != NULL))
		{
		if (ptr != NULL)
			*ptr = 0;

		switch(localuser(msgto, dom))
			{
			case 2:				// mailto local
				msgto = dom->getunknownrelay();
			case 1:				// local user
				if (dolocaldeliver(msgto, filename.get(), dom) != 0)
					addbounce("<%s@%s> unable to deliver local mail.\n", msgto, dom->getname());
				to.remove(n);
				break;
			case 0:
				addbounce("<%s@%s> user unknown.\n", msgto, dom->getname());
				to.remove(n);
				break;
			case -2:			// alias (shouldn't occur at this stage!)
			case -1:			// remote user
				n++;
			break;
			}
		}
	else
		n++;					// remote user
	}
}

void SMTPSERVER::remotedeliver(void)
{
STRING sname;
if (to.getsize() > 0)
	{
	sname = filename.get();

	while(to.getsize() > 50)
		{
		int n;

		writeheader(50);			// write header for first 50 people
		doremotedeliver(sname.get());		// and create output msg

		filename = sessionname();	// get a new session ID

		for(n=0;n<50;n++)			// remove the 50 recipients just processed
			to.remove(0);

		while(delivered.getsize() > 0)		// remove any successfully delivered info from subsequent msgs
			delivered.remove(0);
		}
	if (to.getsize() > 0)				// deliver for final 1-50 users
		{
		writeheader();
		doremotedeliver(sname.get());
		}
	filename = sname.get();				// restore old session ID for cleanup
	}
else if (debug > 2)
	syslog(LOG_MAIL + LOG_DEBUG, "No remote users for delivery");
}

void SMTPSERVER::doremotedeliver(char *sess)
{
FILE *fpin, *fpout;
char buf[512];
int n;

if (sess == NULL)
	sess = filename.get();

fpin = fopen(makefilename(domain->getqin(), 'M', sess), "r+");
fpout = fopen(makefilename(domain->getqout(), 'R', filename.get()), "w");

if (fpin == NULL || fpout == NULL)		// TODO: need to trap this error and report it!
	{
	syslog(LOG_MAIL + LOG_DEBUG, "** Error opening message files for remote deliver");
	return;
	}

lock(fpin); lock(fpout);	
	
while((n = fread(buf,  1, 512, fpin)) > 0)
	fwrite(buf, 1, n, fpout);

// TODO: need to check feof() rather than ferror() has caused fread to finish	
		
unlock(fpin); unlock(fpout);

fclose(fpin);
fclose(fpout);

fpin = fopen(makefilename(domain->getqin(), 'E', sess), "r+");
fpout = fopen(makefilename(domain->getqout(), 'F', filename.get()), "w");

if (fpin == NULL || fpout == NULL)		// TODO: need to trap this error and report it!
	{
	syslog(LOG_MAIL + LOG_DEBUG, "** Error opening envelope files for remote deliver");
	return;
	}

lock(fpin);
lock(fpout);	
	
while((n = fread(buf,  1, 512, fpin)) > 0)
	fwrite(buf, 1, n, fpout);

// TODO: need to check feof() rather than ferror() has caused fread to finish	

unlock(fpin);
unlock(fpout);
		
fclose(fpin);
fclose(fpout);
}

// --------------------- This is the old code





//
// Relaying should be enabled under the following circumstances:
//
// the from username matches one in the relaylist	(user@domain)
// the from domain matches a domain in the relaylist	(@domain)
// the remote host matches a hostname in the relaylist	(domain)
// the remote host matches an ip in the relaylist	(w.x.y.z)
//

int SMTPSERVER::checkthisrelay(char *from, char *rule)
{	
if (strchr(rule, '@') != NULL)		// it's an email relay
	{
	if (*rule == '@')		// anyone from the domain
		{
		if (strcasecmp(rule, strchr(from, '@')) == 0)
			return 1;
		else
			return 0;
		}
	else if (strcasecmp(from, rule) == 0)	// specific
		return 1;
	else
		return 0;
	}
else					// it's a host/domain relay
	{
	struct hostent *host;
	char *hname = "unknown";

	if ((host = gethostbyaddr((char *)&saddr.sin_addr.s_addr, sizeof(saddr.sin_addr.s_addr), AF_INET)) != NULL)
		hname = host->h_name;
	if (strchr(rule, '*') != NULL)	// contains wildcards
		{
		char *pfrom, *pdom;
		pfrom = hname; 		// check the CONNECTED hostname
		pdom = rule;
		while(*pfrom != 0 && *pdom != 0)
			{
			if (*pdom == '*')
				{
				pdom++;
				while(*pfrom != *pdom)
					{
					pfrom++;
					if (*pfrom == 0)
						return 0;
					}
				}
			else if (*pdom == *pfrom)
				{
				pdom++;
				pfrom++;
				}
			else
				return 0;
			}
		if (*pfrom == *pdom)
			return 1;
		else
			return 0;
		}
	else if (strchr(rule, '/') != NULL)	// it's a subnet
		{
		char *subnet, *ptr;
		int bmask;
		subnet = strdup(rule);
		if ((ptr = strchr(subnet, '/')) != NULL)
			{
			int a,b,c,d;
			*ptr = 0;
			bmask = strtol(ptr+1, NULL, 10);
			if (bmask <= 32 && bmask >= 0 && sscanf(subnet, "%d.%d.%d.%d", &a, &b, &c, &d) == 4)
				{
				unsigned long addr1, addr2, mask = 0l;
				addr1 = (a << 24) + (b << 16) + (c << 8) + d;
				addr2 =  ntohl(saddr.sin_addr.s_addr);
				while(bmask-- > 0)
					{
					mask >>= 1;
					mask |= 0x80000000;
					}
				if ((addr1 & mask) == (addr2 & mask))
					return 1;
				}
			}
		return 0;
		}
	else				// it's a host (could be IP or domainname)
		{
		char *rhost;
		rhost = strdup(host->h_name);			// copy the name of remote host
		if ((host = gethostbyname(rule)) != NULL)	// get the hostname specified in the rule
			{
			if (strcasecmp(host->h_name, rhost) == 0)
				{
				free(rhost);
				return 1;
				}
			}
		free(rhost);
		return 0;
		}
	}
}


int SMTPSERVER::checkthislist(char *from, TLIST<STRING> *sublist)
{
TWALK<STRING> list;
//LISTITEM *item;

list.attach(sublist);
if (list.start())
	{
	do	{
		if (!checkthisrelay(from, list.getkey()))
			return 0;
		} while(list.next());
	}

//item = sublist->gethead();
//while(item != NULL)
//	{
//	if (!checkthisrelay(from, (char *)item->getkey()))
//		return 0;
//	item = item->getnext();
//	}
return 1;
}

int SMTPSERVER::checkrelay(char *from)
{
int n = 0;
VIRTUAL_DOMAIN *dom;
char *ptr;
TWALK<TLIST<STRING> > list;
//LISTITEM *item;

ptr = strchr(from, '@');

if (ptr == NULL || (dom = domain->finddomain(ptr+1)) == NULL)
	dom = domain;

list.attach(dom->getrelay());
if (list.start())
	{
	do	{
		if (checkthislist(from, list.get()))
			return 1;
		} while(list.next());
	}
	
//if ((list = dom->getrelay()) != NULL)
//	{
//	item = list->gethead();
//	while(item != NULL)
//		{
//		if (checkthislist(from, (LIST *)item->getdata()))
//			return 1;
//		item = item->getnext();
//		}
//	}
return 0;
}


void SMTPSERVER::process_server(char *data)		// very simple state machine!
{
int cmd;
char word[SMTP_MAX_LINE],*fromptr,*toptr;

fromptr = data;
toptr = word;

if (status == SMTP_STATUS_DATA)
	{
	process_data(data);
	return;
	}

if (SMTP_SESSION::debug > 1)
	syslog(LOG_MAIL + LOG_DEBUG, "<< %s\n", data);

while(*fromptr == ' ')
	fromptr++;

while(*fromptr != ' ' && *fromptr != 0)			// parse out the first token on the command line
	*toptr++ = *fromptr++;
*toptr = 0;

while(*fromptr == ' ')
	fromptr++;

this->line = fromptr;

switch((cmd = lookup(word)))				// lookup the command, some commands work irrespective of state
	{
	case -1:					// not a command at all!
		this->line = data;
		serverstatus(500, this->line);
		return;
	case SMTP_CMD_RSET:
		status = SMTP_STATUS_RSET;
		serverstatus(250);
		closefiles();
		cleanup(NULL);
		reset();
		status = SMTP_STATUS_MAIL;
		return;
	case SMTP_CMD_HELP:				// mainly HELP
		{
		int n = 0;
		sendf("214-Welcome to " PROGRAM " " VERSION "\r\n");
		sendf("214-The following commands are supported:\r\n");
		while(SMTP_SESSION::commands[n] != NULL)
			{
			sendf("214-%s %-20s %s\r\n", SMTP_SESSION::commands[n], SMTP_SESSION::command_arg[n], SMTP_SESSION::command_help[n]);
			n++;
			}
		sendf("214-\r\n");
		sendf("214-To report bugs, please email bugs@virtua-web.co.uk\r\n");
		sendf("214 [End of help]\r\n");
		}
		return;
	case SMTP_CMD_QUIT:
		serverstatus(221);
		closesocket();				// abort session!
		this->status = SMTP_STATUS_ABORT;
		return;
	case SMTP_CMD_VRFY:		// 250 <user>
		{
		TLIST<STRING> *expanded;
		if (SMTP_SESSION::domain != NULL && (expanded = domain->expand(this->line)) != NULL)
			{
			if (strchr(this->line, '@') == NULL)
				sendf("250 <%s@%s>\r\n", this->line, domain->getname());
			else
				sendf("250 <%s>\r\n", this->line);
			delete expanded;
			}
		else
			sendf("550 <%s> unknown\r\n", this->line);
		}
		return;
	case SMTP_CMD_EXPN:		// 550 <user> unknown
		{
		TLIST<STRING> *expanded;
		if (SMTP_SESSION::domain != NULL && (expanded = domain->expand(this->line)) != NULL)
			{
			TWALK<STRING> walk;
			walk.attach(expanded);
			if (walk.start())
				{
				do	{
					sendf("250%c<%s>\r\n", walk.last() ? ' ' : '-', walk.getkey());
					} while(walk.next());
				}
			delete expanded;

//			LISTITEM *item;
//			item = expanded->gethead();
//			while(item != NULL)
//				{
//				sendf("250%c<%s>\r\n", (item->getnext() == NULL) ? ' ' : '-', item->getkey());
//				item = item->getnext();
//				}
//			delete expanded;
			}
		else
			sendf("550 <%s> unknown\r\n", this->line);
		}
		return;
	}

switch(status)
	{
	case SMTP_STATUS_HELO:		/* expecting EHLO or HELO */
		switch(cmd)
			{
			case SMTP_CMD_HELO:
			case SMTP_CMD_EHLO:
				serverstatus(250, this->line);
				status = SMTP_STATUS_MAIL;
				break;
			default:
				serverstatus(503, data);
				break;
			}
		break;
	case SMTP_STATUS_MAIL:
	case SMTP_STATUS_COMPLETE:
		switch(cmd)
			{
			case SMTP_CMD_MAIL:
				if (!parsename(line, "FROM:", word))
					serverstatus(501, line);
				else
					{
					this->closefiles();
					this->cleanup(NULL);
					this->reset();			// new message, so auto reset data
					status = SMTP_STATUS_MAIL;
					if (domain->isspam(word))
						{
						syslog(LOG_MAIL + LOG_NOTICE, "MAIL FROM denied by spam trap (%s)", word);
						sendf("421 MAIL FROM denied access (%s)\r\n", word);
						cleanup(NULL);
						reset();
						status = SMTP_STATUS_MAIL;
						}
					else
						{
						setfrom(word);
						relayok = checkrelay(from.get());
						if (openfiles(sessionname()))		// create a session
							{
							syslog(LOG_MAIL + LOG_NOTICE, "%s: MAIL FROM %s", filename.get(), word);
							addreceive();
							serverstatus(250, word);
							status = SMTP_STATUS_RCPT;
							}
						else
							{
							serverstatus(451);
							closefiles();	
							cleanup(NULL);
							reset();
							status = SMTP_STATUS_MAIL;
							}
						}
					}
				break;
			default:
				serverstatus(503, data);
				break;
			}
		break;
	case SMTP_STATUS_RCPT:
		switch(cmd)
			{
			case SMTP_CMD_RCPT:
				if (!parsename(line, "TO:", word))
					serverstatus(501, line);
				else
					{
					TLIST<STRING> *local = 0;
					if (domain != NULL && (local = domain->expand(word)) != 0)
						{
						TWALK<STRING> walk;
						walk.attach(local);
						if (walk.start())
							{
							do	{
								if(SMTP_SESSION::debug > 2)
									syslog(LOG_MAIL + LOG_DEBUG, "%s -> %s added", word, walk.getkey());
								addto(walk.getkey());
								} while(walk.next());
							}
						
//						LISTITEM *item;
//						item = local->gethead();
//						while(item != NULL)
//							{
//							addto(item->getkey());
//							item = item->getnext();
//							}

						delete local;
						serverstatus(250, word);
						}
					else
						{
						if (this->relayok)
							{
							if (SMTP_SESSION::debug)
								syslog(LOG_MAIL + LOG_NOTICE, "%s: RCPT TO %s accepted", filename.get(), word);
							addto(word);
							serverstatus(250, word);
							}
						else
							{
							syslog(LOG_MAIL + LOG_NOTICE, "%s: RCPT TO %s REJECTED", filename.get(), word);
							serverstatus(550, word);
							}
						}
					}

				break;
			case SMTP_CMD_DATA:
				serverstatus(354);
				status = SMTP_STATUS_DATA;
				break;
			default:
				serverstatus(503, data);
				break;
			}
		break;
	}
}


void SMTPSERVER::serverstatus(int stat)
{
serverstatus(stat, "[none]");
}

void SMTPSERVER::serverstatus(int stat, char *str)
{
switch(stat)
	{
	case SMTP_STATUS_TIMEOUT:
		closefiles();
		cleanup(NULL);
		reset();
		status = SMTP_STATUS_MAIL;
		break;
	case 500:
		sendf("500 Syntax error, command not recognised (%s)\r\n", str);
		break;
	case 501:	
		sendf("501 Syntax error in parameter: %s\r\n", str);
		break;
	case 503:
		sendf("503 Bad sequence of commands\r\n");
		break;
	case 550:
		sendf("550 <%s> relaying denied.\r\n", str);
		break;
	case 451:
		sendf("451 INTERNAL ERROR. MAIL CANNOT BE SENT.\r\n");
		break;
	case 354:
		sendf("354 Start mail input. Enter a period on a line by itself to end.\r\n");
		break;
	case 250:		// acknowledge command
		switch(this->status)
			{
			case SMTP_STATUS_HELO:			// initial connection
				{
				struct hostent *host;
				char *name = "unknown";
				if ((host = gethostbyaddr((char *)&saddr.sin_addr.s_addr, sizeof(saddr.sin_addr.s_addr), AF_INET)) != NULL)
					name = host->h_name;
				if (domain->isspam(str) || domain->isspam(name))
					{
					syslog(LOG_MAIL + LOG_NOTICE, "HELO denied by spam trap %s [%s]", str, name);
					sendf("421 Service not available to %s\r\n", name);
					}
				else
					{
					syslog(LOG_MAIL + LOG_NOTICE, "HELO %s [%s]", str, name);
					sendf("250 Welcome %s [%s].\r\n", str, name);
					}
				}
				break;
			case SMTP_STATUS_RSET:
				sendf("250 Mail session reset.\r\n");
				break;
			case SMTP_STATUS_MAIL:
//				sendf("250 Mail from %s accepted (%s).\r\n", str, (relayok ? "relaying enabled" : "relaying disabled"));
				sendf("250 Mail from %s accepted.\r\n", str);
				break;
			case SMTP_STATUS_RCPT:
				sendf("250 Recipient %s added.\r\n", str);
				break;
			case SMTP_STATUS_DATA:
				sendf("250 %s Message accepted for delivery.\r\n", filename);
				this->status = SMTP_STATUS_COMPLETE;
				break;
			default:
				sendf("250 Request completed, OK.\r\n");
				break;
			}
		break;
	case 220:
		{
		time_t t;
		char tm[128];

		time(&t);
		strftime(tm, sizeof(tm), "%a, %d %b %Y %H:%M:%S %z", gmtime(&t));
		sendf("220 %s SMTP " PROGRAM " " VERSION " %s\r\n", SMTP_SESSION::hostname, tm);
		}
		break;
	case 221:
		sendf("221 %s closing connection. Bye.\r\n", SMTP_SESSION::hostname);
		closesocket();
		break;
	}
}

void SMTPSERVER::deliver(void)
{
SMTPSERVER away;
char *mto;
char *reason;
int n = 0;
while(n < to.getsize() && (mto = to.get(n)) != NULL)
	{
	char *msgto;
	msgto = strdup(mto);
	if (strchr(msgto, '@') != NULL)
		*(strchr(msgto, '@')) = 0;


	if ((reason = domain->isnomail(msgto)) != 0)
		{
		away.from = domain->getpostmaster();
		away.addto(from.get());

		away.bouncing = 1; away.bounce = 0;
		away.filename = away.sessionname();

		away.awaymail("Your message has not been delivered", "Your message has been received but will not be delivered to the recipient %s for the reason given below.\n\n", reason, filename.get(), msgto);

		away.localdeliver();
		away.writeheader();
		away.remotedeliver();

		away.cleanup(NULL);

		to.remove(n);
		}
	else
		{
		if ((reason = domain->isaway(msgto)) != 0)
			{
			away.from = domain->getpostmaster();
			away.addto(from.get());

			away.bouncing = 1; away.bounce = 0;
			away.filename = away.sessionname();

			away.awaymail("Your message has not been delivered", "Your message has been received but the recipient %s is unable to respond for the reason given below.\n\n", reason, filename.get(), msgto);
			away.localdeliver();
			away.writeheader();
			away.remotedeliver();
			}
		n++;
		}
	free(msgto);
	}


localdeliver();
writeheader();
remotedeliver();

if (bounce)
	{
	sendbounce(filename.get());
	}
cleanup(NULL);
}


void SMTPSERVER::awaymail(char *subj, char *reason, char *msg, char *sess, char *who)
{
FILE *msgfile, *transcript, *header;
char *ptr, tm[SMTP_MAX_LINE];
time_t t;
int n;
VIRTUAL_DOMAIN *dom;

time(&t);
strftime(tm, sizeof(tm), "%a, %d %b %Y %H:%M:%S %z", gmtime(&t));

if ((msgfile = fopen(makefilename(domain->getqin(), 'E', filename.get()), "w")) == NULL)
	return;

lock(msgfile);	

//
// Firstly create a simple envelope file
//

fprintf(msgfile, "Date: %s\n", tm);
fprintf(msgfile, "From: %s\n", from.get());
for(n=0;n<to.getsize();n++)
	fprintf(msgfile, "%s %s\n", (n==0) ? "To:" : "", to.get(n));
fprintf(msgfile, "Subject: %s\n", subj);

unlock(msgfile);

fclose(msgfile);

//
// Now create the message body
//

if ((msgfile = fopen(makefilename(domain->getqin(), 'M', filename.get()), "w")) == NULL)
	return;

lock(msgfile);	

fprintf(msgfile, reason, who);

fprintf(msgfile, "    %s\n", msg);

header = fopen(makefilename(domain->getqin(), 'E', sess), "r+");

//	transcript = fopen(makefilename(domain->getqout(), 'B', sess), "r+");
//	header = fopen(makefilename(domain->getqout(), 'F', sess), "r+");

fprintf(msgfile, "\n------ Message Headers ------\n\n");

if (header == NULL)
	fprintf(msgfile, "    [ ERROR: Message Headers lost ]\n");
else
	{
	
	lock(header);
	
	while(fgets(tm, SMTP_MAX_LINE-1, header)!= NULL)
		fprintf(msgfile, "    %s", tm);
		
	unlock(header);		
		
	fclose(header);
	}

unlock(msgfile);	
	
fclose(msgfile);
}

