
//
//
// Project       : Vorpal-Mail
//
// Source        : thash.h
//
// Creator       : Blane Bramble                    Date          : 26th October 2000
// Amended       : Blane Bramble                    Date          : 2nd February 2001
//
// License       : Copyright Virtua-Web 2000-2001, licensed under GNU General Public License (GPL) v2
//
// Definitions for template style hashed class
//
// 1st Feb 2001
// Derived Template Hash class from hash.h and hash.cc

//#include "string.h"
#include "tlist.h"

#ifndef __HASHTABLE__
#define __HASHTABLE__

template <class T> class HASHTABLE {
	private:
		TLIST<T> *table;
		long hash(char *);
		int tablesize;
	public:
		HASHTABLE();
		HASHTABLE(int);
		~HASHTABLE();
		T *find(char *);
		T *add(char *, T *);
//		TLISTITEM<STRING> *find(char *);
//		TLISTITEM<STRING> *add(char *, STRING *);
		int getsize(void) { return tablesize; };
//		TLIST<STRING> *get(int);
		TLIST<T> *get(int);
//		TLISTITEM<STRING> *touchfind(char *);
		T *touchfind(char *);
		void untouch(void);
	};



//
// Private functions
//

template <class T>
long HASHTABLE<T>::hash(char *s)
{
long n = 0;

while(*s != 0)			// usual nonsense hash function. Mathematically probably very flawed.
	{
	n <<= *s & 1;
	n += *s;
	n += (*(s+1) >> 2);
	s++;
	}

return n % tablesize;
}

//
// PUBLIC class functions
//
template <class T>
HASHTABLE<T>::HASHTABLE(int n)
{
tablesize = n;

table = new TLIST<T>[n];
}

template <class T>
HASHTABLE<T>::HASHTABLE()
{
int i;
tablesize = 1013;
table = new TLIST<T>[1013];
}

template <class T>
HASHTABLE<T>::~HASHTABLE()
{
delete[] table;
}

template <class T>
T *HASHTABLE<T>::find(char *key)
{
int h;
h = hash(key);
if (table[h].gethead() == 0)
	return 0;
return table[h].find(key);	//->getdata();
}

template <class T>
T *HASHTABLE<T>::touchfind(char *key)
{
int h;
h = hash(key);
if (table[h].gethead() == 0)
	return 0;
return table[h].touchfind(key);		//->getdata();
}

template <class T>
void HASHTABLE<T>::untouch(void)
{
int h;
for(h=0;h<tablesize;h++)
	table[h].untouch();
}

template <class T>
T *HASHTABLE<T>::add(char *key, T *data)
{
int h;
h = hash(key);
table[h].add(key, data);
return data;
}

template <class T>
TLIST<T> *HASHTABLE<T>::get(int n)
{
return &table[n % tablesize];
}

#endif
