//
//
// Project       : Vorpal-Mail
//
// Source        : tlist.h
//
// Creator       : Blane Bramble                    Date          : 30th December 2000
// Amended       : Blane Bramble                    Date          : 5th March 2001
//
// License       : Copyright Virtua-Web 2000-2001, licensed under GNU General Public License (GPL) v2
//
// Definitions for linked list class
//
// Changelog:
//
// 20th Dec 2000
// Added support for auto-deleting list data.
//
// 30th Dec 2000
// Started amendments for using templates.
//
// 24th Jan 2001
// Finished template code. Added TWALK class to walk through the list.
//
// 2nd Feb 2001
// Added TWALK::last() method to see if at end of list.
//
// 4th Feb 2001
// Added TWALK::del() and code to correctly handle TWALK::next() after a record has been deleted.
//
// 5th March 2001
// Added TLIST::sort() and code to sort a list.

#ifndef __TLISTITEM__
#define __TLISTITEM__

#include "string.h"

template <class T> class TLISTITEM {
	private:
		TLISTITEM *prev, *next;
		STRING key;
		T *data;
		int itouch;
	public:
		TLISTITEM(char *, T *item);
		~TLISTITEM();
		char *getkey(void) { return key.get(); };
		T *getdata(void) { return data; };
		void setdata(T *d);
		TLISTITEM *getnext(void) { return next; };
		TLISTITEM *getprev(void) { return prev; };
		void setnext(TLISTITEM *n) { next = n; };
		void setprev(TLISTITEM *p) { prev = p; };
		void touch(void) { itouch = 1; };
		void untouch(void) { itouch = 0; };
		int touched(void) { return itouch; };
		int sort(void);
	};

template <class T> class TLIST {
	private:
		TLISTITEM<T> *head, *tail;
	public:
		TLIST(void);
		~TLIST();
		TLISTITEM<T> *add(char *, T *);
		T *find(char *);
		T *finddata(T *);
		T *touchfind(char *);
		void sort(void);
		void untouch(void);
		TLISTITEM<T> *gethead(void) { return head; };
		void del(TLISTITEM<T> *);
	};

template <class T> class TWALK {
	private:
		TLIST<T> *list;
		TLISTITEM<T> *current;
		int delfix;			// used to fix call to next() after del()
	public:
		TWALK(void);
		TWALK(TLIST<T> *);
		~TWALK();
		void attach(TLIST<T> *);
		int start(void);
		int next(void);
		int last(void);
		void del(void);
		T *get(void);
		char *getkey(void);
	};


template <class T>
TLISTITEM<T>::TLISTITEM(char *k, T *item)
{
data = item;
key.set(k);
prev = 0;
next = 0;
itouch = 0;
}

template <class T>
TLISTITEM<T>::~TLISTITEM()
{
if (data != 0)
	delete data;
}

template <class T>
int TLISTITEM<T>::sort(void)
{
int sorting;

if (next != 0)
	{
	if (data != 0 && (*data < *(next->data)))
		{
		STRING oldkey;
		T *olddata;

		oldkey = key;
		olddata = data;	

		key = next->key;
		data = next->data;

		next->key = oldkey;
		next->data = olddata;

		return 1;		// something changed
		}
	return next->sort();
	} 
return 0;
}







template <class T>
TLIST<T>::TLIST(void)
{
head = 0;
tail = 0;
}

template <class T>
TLIST<T>::~TLIST()
{
TLISTITEM<T> *chain, *next;
chain = head;
while(chain != 0)
	{
	next = chain->getnext();
	delete chain;
	chain = next;
	}
}

template <class T>
TLISTITEM<T> *TLIST<T>::add(char *k, T *d)
{
TLISTITEM<T> *item;

if (d == 0 || (item = new TLISTITEM<T>(k, d)) == 0)
	return 0;

if (head == 0)		// first item in list!
	{
	head = item;
	tail = item;
	}
else
	{
	if (tail != 0)	// last item exists, so add to list
		{
		tail->setnext(item);
		item->setprev(tail);
		}
	tail = item;		// and make last
	}
return item;
}

template <class T>
T *TLIST<T>::find(char *k)
{
TLISTITEM<T> *item;

item = head;
while(item != 0)
	{
	if (strcasecmp(k, item->getkey()) == 0)
		return item->getdata();
	item = item->getnext();
	}
return 0;
}

template <class T>
T *TLIST<T>::finddata(T *d)
{
TLISTITEM<T> *item;

item = head;
while(item != 0)
	{
	if (d == item->getdata())
		return item->getdata();
	item = item->getnext();
	}
return 0;
}

template <class T>
void TLIST<T>::del(TLISTITEM<T> *i)
{
if (i->getprev() != 0)		// if there is a link before me
	(i->getprev())->setnext(i->getnext());
if (i->getnext() != 0)		// if there is a link after me
	(i->getnext())->setprev(i->getprev());
if (head == i)
	head = i->getnext();
if (tail == i)
	head = i->getprev();
delete i;
}

template <class T>
void TLIST<T>::untouch(void)
{
TLISTITEM<T> *item;

item = head;
while(item != 0)
	{
	item->untouch();
	item = item->getnext();
	}
}

template <class T>
T *TLIST<T>::touchfind(char *s)
{
TLISTITEM<T> *item;

item = head;
while(item != 0)
	{
	if (!item->touched() && strcasecmp(s, item->getkey()) == 0)
		{
		item->touch();
		return item->getdata();
		}
	item = item->getnext();
	}
return 0;
}

template <class T>
void TLIST<T>::sort(void)
{
if (head != 0)
	while(head->sort())
		;
}







template <class T>
TWALK<T>::TWALK(void)
{
list = 0;
current = 0;
delfix = 0;
}

template <class T>
TWALK<T>::TWALK(TLIST<T> *l)
{
list = l;
current = 0;
delfix = 0;
}

template <class T>
TWALK<T>::~TWALK()
{
}

template <class T>
void TWALK<T>::attach(TLIST<T> *l)
{
list = l;
current = 0;
}

template <class T>
int TWALK<T>::start(void)
{
if (list == 0)
	return 0;
if ((current = list->gethead()) == 0)
	return 0;
return 1;
}

template <class T>
int TWALK<T>::next(void)
{
if (list == 0 || current == 0)
	return 0;
if (delfix)			// if delfix is set, the current pointer IS the next list item.
	{
	delfix = 0;		// clear the pointer
	return 1;
	}
if ((current = current->getnext()) == 0)
	return 0;
return 1;
}

template <class T>
T *TWALK<T>::get(void)
{
if (list == 0 || current == 0)
	return 0;
return current->getdata();
}

template <class T>
void TWALK<T>::del(void)
{
T *ditem;
TLISTITEM<T> *dlist;
if (list == 0 || current == 0)
	return;
delfix = 1;

dlist = current;
ditem = current->getdata();

current = current->getnext();
list->del(dlist);
}

template <class T>
char *TWALK<T>::getkey(void)
{
if (list == 0 || current == 0)
	return 0;
return current->getkey();
}

template <class T>
int TWALK<T>::last(void)
{
if (list == 0 || current == 0)
	return 1;
if (current->getnext() == 0)
	return 1;
return 0;
}

#endif

