
//
//
// Project       : Vorpal-Mail
//
// Source        : virtual.cc
//
// Creator       : Blane Bramble                    Date          : 27th October 2000
// Amended       : Blane Bramble                    Date          : 12th March 2001
//
// License       : Copyright Virtua-Web 2000-2001, licensed under GNU General Public License (GPL) v2
//
// virtual.cc provides the classes required to read in a Vorpal-Mail
// configuration file, and to expand and verify user names using
// that data.
//
// 12th March 2001
// Added code to load and check against a spam list.
//
// 27th March 2001
// Added code to load "away" and "nomail" options.

#include <iostream>

#include <stdio.h>
#include <stdlib.h>
#include <ctype.h>
#include <string.h>

#include "mail.h"

//#include "hash.h"
#include "array.h"
#include "string.h"
#include "virtual.h"
#include "smtpdef.h"

char *VIRTUAL_DOMAIN::keywords[] = {
	"domain",
	"smtprelay",
	"masquerade",
	"relay",
	"user",
	"list",
	"virtual",
	"queuein",
	"queueout",
	"unknown",
	"local",
	"scanner",
	"port",
	"bounceafter",
	"spam",
	"bouncereport",
	"nomail",
	"away",
	NULL,
	};

enum {
	VIRTDOM_KEY_DOMAIN = 0,
	VIRTDOM_KEY_SMTPRELAY,
	VIRTDOM_KEY_MASQUERADE,
	VIRTDOM_KEY_RELAY,
	VIRTDOM_KEY_USER,
	VIRTDOM_KEY_LIST,
	VIRTDOM_KEY_VIRTUAL,
	VIRTDOM_KEY_QUEUEIN,
	VIRTDOM_KEY_QUEUEOUT,
	VIRTDOM_KEY_UNKNOWN,
	VIRTDOM_KEY_LOCAL,
	VIRTDOM_KEY_SCANNER,
	VIRTDOM_KEY_PORT,
	VIRTDOM_KEY_BOUNCEAFTER,
	VIRTDOM_KEY_SPAM,
	VIRTDOM_KEY_BOUNCEREPORT,
	VIRTDOM_KEY_NOMAIL,
	VIRTDOM_KEY_AWAY,
	};

//
// Private class functions
//
char *VIRTUAL_DOMAIN::getword(char *from, char *to)
{
int quoted = 0, empty = 1;
*to = 0;
while(isspace(*from))
	from++;

while(*from != 0)
	{
	if (quoted)
		{
		if (*from == '"')
			{
			quoted = 0;
			from++;
			}
		else
			*to++ = *from++;
		}
	else
		{
		if (isspace(*from))
			break;
		else if (*from == '#')
			{
			from += strlen(from);
			break;
			}
		else if (*from == '"')
			{
			quoted = 1;
			from++;
			}
		else if (*from == ',' && !empty)
			break;
		else
			*to++ = *from++;
		empty = 0;
		}
	}
*to = 0;
return from;
}

int VIRTUAL_DOMAIN::lookup(char *word)
{
int n = 0;
while(keywords[n] != NULL)
	{
	if (strcasecmp(word, keywords[n]) == 0)
		return n;
	n++;
	}
return -1;
}

VIRTUAL_DOMAIN *VIRTUAL_DOMAIN::finddomain(char *dom)
{
int n;
TWALK<VIRTUAL_DOMAIN> list;
if (*dom == 0 || domainname.comparecase(dom) == 0)
	return this;
for(n=0;n<masquerade.getsize();n++)
	{
	if (strcasecmp(masquerade.get(n), dom) == 0)
		return this;
	}
list.attach(&virtualdomain);
if (list.start())
	{
	do	{
		VIRTUAL_DOMAIN *d;
		if ((d = list.get())->finddomain(dom) != 0)
			return d;
		} while(list.next());
	}
//if (virtualdomain != NULL)
//	{
//	LISTITEM *item;
//	item = virtualdomain->gethead();
//	while(item != NULL)
//		{
//		VIRTUAL_DOMAIN *d;
//		if ((d = ((VIRTUAL_DOMAIN *)(item->getdata()))->finddomain(dom)) != NULL)
//			return d;
//		item = item->getnext();
//		}
//	}
return 0;
}

VIRTUAL_DOMAIN::VIRTUAL_DOMAIN(void)
{
int n;

listenport = 25;
// relaylist = new LIST();
status = VIRTDOM_KEYWORD;
scratch = NULL;
queuein.set("/var/spool/vmail/in");
queueout.set("/var/spool/vmail/out");
unknown = VIRTDOM_UNKNOWN_REJECT;
bounceafter = (60 * 60 * 24 * 5);		// default to 5 day bounce

for(n=0;n<MAX_BOUNCE_REPORT;n++)
	bouncearray[n] = -1;
}

VIRTUAL_DOMAIN::~VIRTUAL_DOMAIN()
{
//if (relaylist != NULL)
//	{
//	LISTITEM *item;
//	item = relaylist->gethead();
//	while(item != NULL)
//		{
//		delete ((LIST *)(item->getdata()));
//		item = item->getnext();
//		}
//	delete relaylist;
//	}
if (scratch != NULL)
	delete scratch;
}


void VIRTUAL_DOMAIN::addbouncereport(int n)
{
int i;
for(i=0;i<MAX_BOUNCE_REPORT;i++)
	{
	if (bouncearray[i] == -1)
		{
		bouncearray[i] = n;
		return;
		}
	}
}

int VIRTUAL_DOMAIN::checkbouncecount(long diff, int n)
{
long t;
if (n < MAX_BOUNCE_REPORT && bouncearray[n] != -1)
	{
	t = bouncearray[n] * (24 * 60 * 60);
	if (diff >= t)
		return n++;
	}
return n;
}

void VIRTUAL_DOMAIN::process(char *w)
{
static STRING thiskey;

switch(this->status)
	{
	case VIRTDOM_KEYWORD:
		if (*w == ',')
			status = VIRTDOM_VALUE;			// it's another value!
		else if ((command = lookup(w)) == -1)
			status = VIRTDOM_KEYWORD;		// unknown keyword
		else
			status = VIRTDOM_ARGUMENT;
		break;
	case VIRTDOM_ARGUMENT:
		switch(command)
			{
			case VIRTDOM_KEY_DOMAIN:
				if (domainname.get() == 0)
					{
					char tmp[SMTP_MAX_NAME];
					snprintf(tmp, SMTP_MAX_NAME-1, "postmaster@%s", w);
					postmaster = tmp;
					domainname = w;
					}
				status = VIRTDOM_KEYWORD;
				break;
			case VIRTDOM_KEY_SCANNER:
				if (scanner.get() == 0)
					scanner.set(w);
				status = VIRTDOM_KEYWORD;
				break;
			case VIRTDOM_KEY_PORT:
				listenport = (int)strtol(w, NULL, 10);
				status = VIRTDOM_KEYWORD;
				break;
			case VIRTDOM_KEY_SMTPRELAY:
				if (smtprelay.get() == 0)
					smtprelay.set(w);
				status = VIRTDOM_KEYWORD;
				break;
			case VIRTDOM_KEY_BOUNCEAFTER:
				bounceafter = (strtol(w, NULL, 10) * 60 * 60 * 24);		// convert days to seconds!
				status = VIRTDOM_KEYWORD;
				break;
			case VIRTDOM_KEY_MASQUERADE:
				masquerade.add(w);
				status = VIRTDOM_KEYWORD;
				break;
			case VIRTDOM_KEY_SPAM:
				spamlist.add(NULL, new STRING(w));
				status = VIRTDOM_KEYWORD;
				break;
			case VIRTDOM_KEY_BOUNCEREPORT:
				addbouncereport(strtol(w, NULL, 10));
				status = VIRTDOM_KEYWORD;
				break;
			case VIRTDOM_KEY_RELAY:
//				{
//				LIST *data;
//
//				if ((data = new LIST()) != NULL)
//					{
//					data->add(w, "relay");			// add relay item to list
//	
//					relaylist->add(NULL, data);		// add new list to relaylist
//	
//					sublist = data;
//					}
//				status = VIRTDOM_KEYWORD;
//				}
//				break;


				{
				TLIST<STRING> *data;
				if ((data = new TLIST<STRING>()) != 0)
					{
					data->add(w, new STRING("relay"));
					relaylist.add(0, data);
					sublist = data;
					}
				status = VIRTDOM_KEYWORD;
				}
				break;

			case VIRTDOM_KEY_VIRTUAL:
				{
				VIRTUAL_DOMAIN *dom;
				dom = new VIRTUAL_DOMAIN();
				virtualdomain.add(w, dom);
				dom->load(w);
				status = VIRTDOM_KEYWORD;
				}
				break;
			case VIRTDOM_KEY_NOMAIL:
				thiskey = w;
				status = VIRTDOM_VALUE;
				break;
			case VIRTDOM_KEY_AWAY:
				thiskey = w;
				status = VIRTDOM_VALUE;
				break;
			case VIRTDOM_KEY_USER:
				thiskey = w;
				if (userlist.find(w) == NULL)
					userlist.add(w, new TLIST<STRING>);
				status = VIRTDOM_VALUE;
				break;
			case VIRTDOM_KEY_LIST:
				thiskey = w;
				if (listlist.find(w) == 0)
					listlist.add(w, new TLIST<STRING>);
				status = VIRTDOM_VALUE;
				break;
			case VIRTDOM_KEY_QUEUEIN:
				queuein.set(w);
				status = VIRTDOM_KEYWORD;
				break;
			case VIRTDOM_KEY_QUEUEOUT:
				queueout.set(w);
				status = VIRTDOM_KEYWORD;
				break;
			case VIRTDOM_KEY_UNKNOWN:
				if (strcasecmp(w, "reject") == 0)
					unknown = VIRTDOM_UNKNOWN_REJECT;
				else if (strcasecmp(w, "accept") == 0)
					unknown = VIRTDOM_UNKNOWN_ACCEPT;
				else if (strcasecmp(w, "relay") == 0)
					unknown = VIRTDOM_UNKNOWN_RELAY;
				else
					{
					unknownrelay.set(w);
					unknown = VIRTDOM_UNKNOWN_MAILTO;
					}
				status = VIRTDOM_KEYWORD;
				break;
			case VIRTDOM_KEY_LOCAL:
				local.set(w);
				status = VIRTDOM_KEYWORD;
				break;
			}
		break;
	case VIRTDOM_VALUE:
		switch(command)
			{
			TLIST<STRING> *item;
//			LISTITEM *item;
//			LIST *data;
			char *ptr;

			case VIRTDOM_KEY_NOMAIL:
				nomaillist.add(thiskey.get(), new STRING(w));
				status = VIRTDOM_KEYWORD;
				break;
			case VIRTDOM_KEY_AWAY:
				awaylist.add(thiskey.get(), new STRING(w));
				status = VIRTDOM_KEYWORD;
				break;
			case VIRTDOM_KEY_RELAY:
				sublist->add(w, new STRING("relay"));
				status = VIRTDOM_KEYWORD;
				break;

			case VIRTDOM_KEY_BOUNCEREPORT:
				addbouncereport(strtol(w, NULL, 10));
				status = VIRTDOM_KEYWORD;
				break;
			case VIRTDOM_KEY_USER:
#ifdef OLDMETHOD
				if ((ptr = strchr(w, ',')) != NULL)		// if ends in , (more values expected)
					*ptr = 0;				// remove ,
				else
					status = VIRTDOM_KEYWORD;		// otherwise expect a keyword next
#else
				status = VIRTDOM_KEYWORD;
#endif
				if ((item = userlist.find(thiskey.get())) != 0)
					item->add(w, new STRING("user"));

//				if ((item = userlist.find(thiskey.get())) != NULL)
//					{
//					data = (LIST *)(item->getdata());
//					data->add(w, "user");
//					}
				break;
			case VIRTDOM_KEY_LIST:
#ifdef OLDMETHOD
				if ((ptr = strchr(w, ',')) != NULL)		// if ends in , (more values expected)
					*ptr = 0;				// remove ,
				else
					status = VIRTDOM_KEYWORD;		// otherwise expect a keyword next
#else
				status = VIRTDOM_KEYWORD;
#endif
				if ((item = listlist.find(thiskey.get())) != 0)
					item->add(w, new STRING("list"));

//				if ((item = listlist.find(thiskey.get())) != NULL)
//					{
//					data = (LIST *)(item->getdata());
//					data->add(w, "user");
//					}
				break;
			}
		break;
	}
}

int VIRTUAL_DOMAIN::load(char *fn)
{
FILE *fp;
char linebuf[MAX_LINE];
char *ptr;
char wordbuf[MAX_LINE];

ptr = linebuf;
*ptr = 0;

if ((fp = fopen(fn, "r")) == NULL)
	return 0;

while(fgets(linebuf, MAX_LINE-1, fp) != NULL)
	{
	ptr = linebuf;
	while(*(ptr = getword(ptr, wordbuf)) != 0)
		{
		process(wordbuf);
		}
	}
fclose(fp);
}

void VIRTUAL_DOMAIN::dump(void)
{
int n;
char *str;
//LIST *list, *data;
TWALK<TLIST<STRING> > listwalk;
TWALK<STRING> subwalk;
TWALK<VIRTUAL_DOMAIN> domwalk;
TLIST<TLIST<STRING> > *tlist;

std::cout << "---- DUMP ----" << endl;

if (smtprelay.get() != 0)
	std::cout <<  "smtprelay = " << smtprelay.get() << endl;
if (domainname.get() != 0)
	std::cout << "domain = " << domainname.get() << endl;
if (local.get() != 0)
	std::cout << "local = " << local.get() << endl;	

if (scanner.get() != 0)
	std::cout << "scanner = " << scanner.get() << endl;

std::cout << "unknown = " << unknown << endl;

n = 0;
while((str = masquerade.get(n)) != NULL)
	{
	std::cout << "masquerade[" << n << "] = " << str << endl;
	n++;
	}

subwalk.attach(&spamlist);
if (subwalk.start())
	{
	do	{
		std::cout << "spam " << subwalk.get()->get() << endl;
		} while(subwalk.next());
	}



listwalk.attach(&relaylist);
if (listwalk.start())
	{
	do	{
		subwalk.attach(listwalk.get());
		if (subwalk.start())
			{
			std::cout << "relay ";
			do	{
				std::cout << subwalk.getkey() << ", ";
				} while(subwalk.next());
			std::cout << endl;
			}
		} while(listwalk.next());
	}

//	{
//	LISTITEM *item;	
//
//	item = relaylist->gethead();
//	while(item != NULL)
//		{
//		LISTITEM *ritem;
//		data = ((LIST *)item->getdata());
//		if (data != NULL)
//			{
//			ritem = data->gethead();
//			std::cout << "relay ";
//			while(ritem != NULL)
//				{
//				std::cout << ritem->getkey(), ", ";
//				ritem = ritem->getnext();
//				}
//			}
//		std::cout << endl;
//		item = item->getnext();
//		}
//	}

for(n=0;n<userlist.getsize();n++)
	{
	if ((tlist = userlist.get(n)) != NULL)		// get the list of lists from the hash table
		{
		TWALK<TLIST<STRING> > listwalk;
		listwalk.attach(tlist);
		if (listwalk.start())
			{
			do	{
				TWALK<STRING> stringwalk;
				std::cout << listwalk.getkey() << " = ";
				stringwalk.attach(listwalk.get());
				if (stringwalk.start())
					{
					do	{
						std::cout << stringwalk.getkey() << ", ";
						} while(stringwalk.next());
					}
				std::cout << endl;
				} while(listwalk.next());
			}
		}
	}


//		LISTITEM *item;
//		item = list->gethead();			// get the first list with this hash value
//		while(item != NULL)			// for each hash list item (primary name)
//			{
//			LISTITEM *ditem;
//			std::cout << item->getkey() << " = ";
//			if ((data = (LIST *)item->getdata()) != NULL)		// get the list of secondary names
//				{
//				ditem = data->gethead();
//				while(ditem != NULL)
//					{
//					std::cout << ditem->getkey() << ", ";
//					ditem = ditem->getnext();
//					}
//				}
//			std::cout << endl;
//			item = item->getnext();
//			}
//		}
//	}
	
for(n=0;n<listlist.getsize();n++)
	{
	if ((tlist = listlist.get(n)) != 0)		// get the list of lists from the hash table
		{
		TWALK<TLIST<STRING> > listwalk;
		listwalk.attach(tlist);
		if (listwalk.start())
			{
			do	{
				TWALK<STRING> stringwalk;
				std::cout << listwalk.getkey() << " = ";
				stringwalk.attach(listwalk.get());
				if (stringwalk.start())
					{
					do	{
						std::cout << stringwalk.getkey() << ", ";
						} while(stringwalk.next());
					}
				std::cout << endl;
				} while(listwalk.next());
			}
		}
	}

//		LISTITEM *item;
//		item = list->gethead();			// get the first list with this hash value
//		while(item != NULL)			// for each hash list item (primary name)
//			{
//			LISTITEM *ditem;
//			std::cout << item->getkey() << " = ";
//			if ((data = (LIST *)item->getdata()) != NULL)		// get the list of secondary names
//				{
//				ditem = data->gethead();
//				while(ditem != NULL)
//					{
//					std::cout << ditem->getkey() << ", ";
//					ditem = ditem->getnext();
//					}
//				}
//			std::cout << endl;
//			item = item->getnext();
//			}
//		}
//	}

domwalk.attach(&virtualdomain);
if (domwalk.start())
	{
	do	{
		VIRTUAL_DOMAIN *dom;
		dom = domwalk.get();
		std::cout << domwalk.getkey() << ":" << endl;
		dom->dump();
		} while(domwalk.next());
	}
}

char *VIRTUAL_DOMAIN::toemail(char *from)
{
static char full[SMTP_MAX_NAME];
if (strchr(from, '@') != NULL)
	return from;
snprintf(full, SMTP_MAX_NAME - 1,"%s@%s", from, this->domainname.get());
return full;
}

int VIRTUAL_DOMAIN::isspam(char *from)
{
TWALK<STRING> walk;

if (strchr(from, '@') != NULL)
	from = strchr(from, '@') + 1;

walk.attach(&spamlist);
if (walk.start())
	{
	do	{
		char *test,*spam;
		spam = walk.get()->get();
		test = from;

		while(*test != 0 && *spam != 0)
			{
			if (*spam == '*')
				{
				spam++;
				while(*test != 0 && *spam != 0 && (*test != *spam))
					test++;
				}
			else if (*test == *spam)
				{
				test++;
				spam++;
				}
			else
				break;
			}
		if (*test == 0 &&  *spam == 0)
			return 1;
		} while(walk.next());
	}
return 0;
}


TLIST<STRING> *VIRTUAL_DOMAIN::expand(char *email)
{
int hasdomain = 0;
char *domain, *name, *temp, *ptr;
VIRTUAL_DOMAIN *dom;
TLIST<STRING> *expanded = 0, *list;
TLIST<STRING> *item;
//LIST *expanded = NULL, *list;
//LISTITEM *item;

temp = strdup(email);

ptr = temp;
while(*ptr != 0)
	{
	if (*ptr == '@')
		{
		*ptr = 0;
		hasdomain = 1;
		}
	else
		ptr++;
	}
name = strdup(temp);
if (hasdomain)
	{
	*ptr = '@';
	domain = strdup(ptr+1);
	}
else
	{
	domain = strdup(this->getname());
	}
free(temp);

// At this point, name and domain are valid.

if ((dom = finddomain(domain)) == NULL)
	{
	free(name);
	free(domain);
	return NULL;
	}

listlist.untouch();
userlist.untouch();

if ((item = dom->listlist.touchfind(name)) != 0)
	{
	TWALK<STRING> walk;
	walk.attach(item);
	if (walk.start())
		{
		TLIST<STRING> *recurse;
		do	{
			if (expanded == 0)
				expanded = new TLIST<STRING>;
			if ((recurse = expand(dom->toemail(walk.getkey()))) != 0)		// an item was a list!
				{
				TWALK<STRING> rwalk;
				rwalk.attach(recurse);
				if (rwalk.start())
					{
					do	{
						expanded->add(dom->toemail(rwalk.getkey()), new STRING(rwalk.get()));
						} while(rwalk.next());
					}
				delete recurse;
				}
			else
				expanded->add(dom->toemail(walk.getkey()), new STRING(walk.get()));
			} while(walk.next());
		}
	}
else if ((item = dom->userlist.touchfind(name)) != 0)
	{
	TWALK<STRING> walk;
	walk.attach(item);
	if (walk.start())
		{
		TLIST<STRING> *recurse;
		do	{
			if (expanded == 0)
				expanded = new TLIST<STRING>;
#ifdef EXPAND_USERS
			if ((recurse = expand(dom->toemail(walk.getkey()))) != 0)		// an item was a list!
				{
				TWALK<STRING> rwalk;
				rwalk.attach(recurse);
				if (rwalk.start())
					{
					do	{
						expanded->add(dom->toemail(rwalk.getkey()), new STRING(rwalk.get()));
						} while(rwalk.next());
					}
				delete recurse;
				}
			else
#endif
				expanded->add(dom->toemail(walk.getkey()), new STRING(walk.get()));
			} while(walk.next());
		}
	}


#ifdef OLDMETHOD
if ((item = dom->listlist.touchfind(name)) != NULL)		// got a name in an expansion list
	{
	if ((list = (LIST *)item->getdata()) != NULL)	// get the list of expansions (shouldn't ever be null, but what the hell!
		{
		item = list->gethead();
		while(item != NULL)			// walk the valid expansions
			{
			LIST *recurse;
			if (expanded == NULL)
				expanded = new LIST;
			if ((recurse = expand(dom->toemail(item->getkey()))) != NULL)		// an item was a list!
				{
				LISTITEM *ritem;
				ritem = recurse->gethead();
				while(ritem != NULL)
					{
					expanded->add(dom->toemail(ritem->getkey()), new STRING(ritem->getdata()));	
					ritem = ritem->getnext();
					}
				delete recurse;
				}
			else
				expanded->add(dom->toemail(item->getkey()), new STRING(item->getdata()));
			item = item->getnext();
			}
		}
	}
else if ((item = dom->userlist.touchfind(name)) != NULL)		// got a name in an expansion list
	{
	if ((list = (LIST *)item->getdata()) != NULL)	// get the list of users.
		{
		item = list->gethead();
		while(item != NULL)			// walk the valid expansions
			{
			LIST *recurse;
			if (expanded == NULL)
				expanded = new LIST();
#ifdef EXPAND_USERS
			if ((recurse = expand(dom->toemail(item->getkey()))) != NULL)		// expand
				{
				LISTITEM *ritem;
				ritem = recurse->gethead();
				while(ritem != NULL)
					{
					expanded->add(dom->toemail(ritem->getkey()), new STRING(ritem->getdata()));
					ritem = ritem->getnext();
					}
				delete recurse;
				}
			else
#endif
				expanded->add(dom->toemail(item->getkey()), new STRING(item->getdata()));
			item = item->getnext();
			}
		}
	}
#endif

free(name);
free(domain);
return expanded;
}

char *VIRTUAL_DOMAIN::isnomail(char *who)
{
STRING *msg;
if ((msg = nomaillist.find(who)) != 0)
	return msg->get();
else
	return 0;
}

char *VIRTUAL_DOMAIN::isaway(char *who)
{
STRING *msg;
if ((msg = awaylist.find(who)) != 0)
	return msg->get();
else
	return 0;
}

/* User verify method:

Split into user and domain
Lookup domain in list of virtual domains, if found use that .conf, otherwise use vmail.conf (defaults)
Lookup user in list
If found, expand and lookup each expansion in list (recursively)
If not found, lookup user in user
If found, expand, attempt local delivery for any without domain. Any with domain, lookup domain in virtual domains, if found start again, else queue for remote delivery.


*/

