
//
//
// Project       : Vorpal-Mail
//
// Source        : virtual.h
//
// Creator       : Blane Bramble                    Date          : 27th October 2000
// Amended       : Blane Bramble                    Date          : 27th March 2001
//
// License       : Copyright Virtua-Web 2000-2001, licensed under GNU General Public License (GPL) v2
//
// Definitions for virtual domain configurations for Vorpal-Mail
//
//

//#include "hash.h"
#include "array.h"

#ifndef __VIRTUAL_DOMAIN__
#define __VIRTUAL_DOMAIN__

#include "string.h"
#include "tlist.h"
#include "thash.h"

#define EXPAND_USERS

#define MAX_BOUNCE_REPORT 5

enum {
	VIRTDOM_KEYWORD,
	VIRTDOM_ARGUMENT,
	VIRTDOM_VALUE
	};

enum VIRTDOM_UNKNOWN {
	VIRTDOM_UNKNOWN_REJECT,
	VIRTDOM_UNKNOWN_RELAY,
	VIRTDOM_UNKNOWN_ACCEPT,
	VIRTDOM_UNKNOWN_MAILTO,
	};

class VIRTUAL_DOMAIN {
	private:
		static char *keywords[];

		int command, status;

		char *scratch;

		STRING queuein, queueout;
		STRING domainname, smtprelay, local;
		STRING scanner, unknownrelay;
		STRING postmaster;
		int listenport;
		long bounceafter;

		int bouncearray[MAX_BOUNCE_REPORT];
		STRING_ARRAY masquerade;
		TLIST<TLIST<STRING> > relaylist;
		TLIST<STRING> spamlist;
		TLIST<STRING> awaylist;
		TLIST<STRING> nomaillist;
		TLIST<STRING> *sublist;
		HASHTABLE<TLIST<STRING> > userlist;
		HASHTABLE<TLIST<STRING> > listlist;
//		HASHTABLE userlist;
//		HASHTABLE listlist;
		TLIST<VIRTUAL_DOMAIN> virtualdomain;

		VIRTDOM_UNKNOWN unknown;

		char *getword(char *, char *);
		int lookup(char *);
		void process(char *);
		void addbouncereport(int);
	public:
		VIRTUAL_DOMAIN(void);
		~VIRTUAL_DOMAIN();
		int load(char *);
		void dump(void);
		TLIST<VIRTUAL_DOMAIN> *getvirtual(void) { return &virtualdomain; };
		TLIST<TLIST<STRING> > *getrelay(void) { return &relaylist; };
		char *getsmtp(void) { return smtprelay.get(); };
		TLIST<STRING> *expand(char *);
		char *toemail(char *);
		char *getname(void) { return domainname.get(); };
		char *getqin(void) { return queuein.get(); };
		char *getqout(void) { return queueout.get(); };
		VIRTUAL_DOMAIN *finddomain(char *);
		VIRTDOM_UNKNOWN getunknown(void) { return unknown; };
		char *getunknownrelay(void) { return unknownrelay.get(); };
		char *getlocal(void) { return local.get(); };
		char *getscanner(void) { return scanner.get(); };
		int getport(void) { return listenport; };
		long getbounceafter(void) { return bounceafter; };
		int isspam(char *);
		int checkbouncecount(long, int);
		char *isnomail(char *);
		char *isaway(char *);
		char *getpostmaster(void) { return postmaster.get(); };
	};

#endif

//
// userlist [HASHTABLE]
//  |
// LIST
//  |
// LISTITEM (user1) -- LIST
//  |                   |
//  |                  LISTITEM (alias1)
//  |                  LISTITEM (alias2)
//  |
// LISTITEM (user2) -- LIST
//                      |
//                     LISTITEM (alias1)
//                     LISTITEM (alias2)

